# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Salesforce Instance.

SalesforceInstance is the only class defined in this file.

SalesforceeInstance:    Derived class from CloudAppsInstance Base class, representing a
                            Salesforce instance, and to perform operations on that instance

SalesforceInstance:

    _restore_json()                     --  Returns the JSON request to pass to the API as per the options selected by
                                            the user

    restore_to_file_system()            --  Runs object level restore to file system and returns object of Job or
                                            Schedule class

    restore_to_database()               --  Runs object level restore to database and returns object of Job or Schedule
                                            class

    restore_to_salesforce_from_database() --  Runs restore to Salesforce from database and returns object of Job or
                                            Schedule class

    restore_to_salesforce_from_media()  --  Runs restore to Salesforce from database and returns object of Job or
                                            Schedule class

    metadata_restore_to_file_system()   --  Runs metadata restore to file system and returns object of Job or Schedule
                                            class

    metadata_restore_to_salesforce()    --  Runs metadata restore to Salesforce and returns object of Job or Schedule
                                            class

SalesforceInstance Attributes:

    **ca_instance_type**            --  Returns the instance type of this cloud apps instance

    **organization_id**             --  Returns the Salesforce organization id

    **login_url**                   --  Returns the login url of Salesforce organization

    **consumer_id**                 --  Returns the Consumer Id of the Salesforce connected app used to authenticate
                                        with Salesforce by this instance

    **proxy_client**                --  Returns the name of the access node. Returns None if client group is configured
                                        as access node

"""
from __future__ import unicode_literals
from base64 import b64encode
from ..cainstance import CloudAppsInstance
from ...exception import SDKException


class SalesforceInstance(CloudAppsInstance):
    """Class for representing an Instance of the Salesforce instance type."""

    @property
    def ca_instance_type(self):
        """
        Returns the instance type of this cloud apps instance

        Returns:
            (str): Instance Type
        """
        return 'SALESFORCE'

    @property
    def organization_id(self):
        """
        Returns the Salesforce organization id

        Returns:
            (str): Organization Id

        Raises:
            SDKException: if attribute could not be fetched
        """
        try:
            return self._properties['cloudAppsInstance']['salesforceInstance']['sfOrgID']
        except KeyError:
            raise SDKException('Instance', '105', 'Could not fetch organization ID')

    @property
    def login_url(self):
        """
        Returns the login url of Salesforce organization

        Returns:
            (str): Login URL

        Raises:
            SDKException: if attribute could not be fetched
        """
        try:
            return self._properties['cloudAppsInstance']['salesforceInstance']['endpoint']
        except KeyError:
            raise SDKException('Instance', '105', 'Could not fetch login url')

    @property
    def consumer_id(self):
        """
        Returns the Consumer Id of the Salesforce connected app used to authenticate with Salesforce by this instance

        Returns:
            (str): Consumer Id

        Raises:
            SDKException: if attribute could not be fetched
        """
        try:
            return self._properties['cloudAppsInstance']['salesforceInstance']['consumerId']
        except KeyError:
            raise SDKException('Instance', '105', 'Could not fetch login url')

    @property
    def proxy_client(self):
        """
        Returns the name of the access node.

        Returns:
            (str): Access Node

        Raises:
            SDKException:
                if attribute could not be fetched

                if access node is a client group
        """
        try:
            general_cloud_properties = self._properties['cloudAppsInstance']['generalCloudProperties']
            if 'clientName' in general_cloud_properties['proxyServers'][0].keys():
                return general_cloud_properties['proxyServers'][0]['clientName']
            if 'clientName' in general_cloud_properties['accessNodes']['memberServers'][0]['client'].keys():
                return general_cloud_properties['accessNodes']['memberServers'][0]['client']['clientName']
            if 'clientGroupName' in general_cloud_properties['accessNodes']['memberServers'][0]['client'].keys():
                raise SDKException(
                    'Instance',
                    '102',
                    'This instance uses a client group as access node. Use access_node attribute instead.'
                )
        except KeyError:
            raise SDKException('Instance', '105', 'Could not fetch proxy client')

    @property
    def access_node(self):
        """
        Returns a dictionary containing clientName and clientId or clientGroupName and clientGroupId depending on
        whether a single client or a client group is configured as access node.

        Returns:
            (dict): Dictionary containing access node name and id

        Raises:
            SDKException: if attribute could not be fetched
        """
        try:
            access_node = self._properties['cloudAppsInstance']['generalCloudProperties']['accessNodes'] \
                ['memberServers'][0]['client'].copy()
            if 'entityInfo' in access_node:
                del access_node['entityInfo']
            return access_node
        except KeyError:
            raise SDKException('Instance', '105', 'Could not fetch access node')

    def _restore_json(self, **kwargs):
        """
        Returns the JSON request to pass to the API as per the options selected by the user

        Args:
            **kwargs (dict): Dict of named parameters to set for restore

        Returns:
            (dict): Request JSON
        """
        if len(self.backupsets.all_backupsets) > 1 or len(self.subclients.all_subclients) > 1:
            raise SDKException(
                'Instance',
                '102',
                'More than one backupset/subclient configured in this instance. Run restore from subclient'
            )

        if not kwargs.get('no_of_streams', None):
            kwargs['no_of_streams'] = 2
        kwargs['client'] = kwargs.get('client', None) or self._agent_object._client_object
        request_json = super()._restore_json(**kwargs)

        backupset = self.backupsets.get(list(self.backupsets.all_backupsets.keys())[0])
        subclient = list(self.subclients.all_subclients.items())[0]
        request_json['taskInfo']['associations'][0].update({
            'backupsetName': backupset.name,
            'backupsetId': int(backupset.backupset_id),
            'subclientName': subclient[0],
            'subclientId': int(subclient[1]['id'])
        })

        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['cloudAppsRestoreOptions'] = {
            'instanceType': self.ca_instance_type,
            'salesforceRestoreOptions': {
                'restoreToFileSystem': kwargs.get('restore_to_file_system', False),
                'restoreToSalesforce': kwargs.get('restore_to_salesforce', False),
                'restoreFromDatabase': kwargs.get('restore_from_database', False),
                'isMetadataRestore': kwargs.get('is_metadata_restore', False),
                'pathToStoreCsv': kwargs.get('path_to_store_csv', None) or backupset.download_cache_path,
                'dependentRestoreLevel': kwargs.get('dependent_restore_level', 0),
                'restoreParentType': kwargs.get('restore_parent_type', 'NONE'),
                'isSaaSRestore': False
            }
        }

        if 'restore_to_salesforce' in kwargs and kwargs['restore_to_salesforce']:
            if kwargs.get('instance', None) and kwargs.get('backupset', None):
                destination_client = self._commcell_object.clients.get(kwargs['client'])
                destination_instance = destination_client.agents.get('Cloud Apps').instances.get(kwargs['instance'])
                destination_backupset = destination_instance.backupsets.get(kwargs['backupset'])
            else:
                destination_instance = self
                destination_backupset = backupset
            request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['destination'].update({
                'destinationInstance': {
                    'instanceId': int(destination_instance.instance_id),
                    'instanceName': destination_instance.name,
                },
                'destinationBackupset': {
                    'backupsetId': int(destination_backupset.backupset_id),
                    'backupsetName': destination_backupset.backupset_name
                }
            })

        if kwargs.get('db_enabled', False):
            request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['cloudAppsRestoreOptions'] \
                ['salesforceRestoreOptions'].update({
                'syncDatabase': {
                    'dbEnabled': True,
                    'dbType': kwargs['db_type'],
                    'dbHost': kwargs['db_host'],
                    'dbInstance': kwargs.get('db_instance', ''),
                    'dbPort': str(kwargs.get('db_port', 5432 if kwargs['db_type'] == 'POSTGRESQL' else 1433)),
                    'dbName': kwargs['db_name'],
                    'dbUserPassword': {
                        'userName': kwargs['db_user_name'],
                        'password': b64encode(kwargs['db_password'].encode()).decode()
                    }
                },
                'overrideTable': kwargs.get('override_table', True),
                'restoreCatalogDatabase': kwargs.get('restore_catalog_database', False)
            })

        return request_json

    def restore_to_file_system(self, **kwargs):
        """
        Runs object level restore to file system and returns object of Job or Schedule class. For out of place restore,
        pass both client and path_to_store_csv parameters. By default, will restore to access node and download cache
        path.

        Args:
            **kwargs (dict): Restore options including
                {
                    paths (list[str]): List of files and objects to restore like
                                ['/Files/filename', '/Objects/object_name']
                                (Default is ['/Files/', '/Objects/'] which selects all files and objects for restore),

                    client (str): Name of destination client (Default is access node),

                    path_to_store_csv (str): path on destination client to restore to (Default is download cache path),

                    from_time (str): time to restore contents after like YYYY-MM-DD HH:MM:SS (Default is None),

                    to_time (str): time to restore contents before like YYYY-MM-DD HH:MM:SS (Default is None),

                    no_of_streams (int): Number of streams to use for restore (Default is 2),

                    dependent_restore_level (int): restore children option (Default is 0)
                                                    0  -- No Children
                                                    1  -- Immediate Children
                                                    -1 -- All Children,

                    restore_parent_type (str): restore parents option (Default is 'NONE')
                                                    'NONE' -- No Parents
                                                    'ALL'  -- All Parents
                }

        Returns:
            object: Object of Job or Schedule class

        Raises:
            SDKException:
                if paths is given but is not a list
    
                if client parameter is not given and the access node configured with this instance is a client group
                
                if either client or path_to_store_csv is given but not both are present
                
                if client or path_to_store_csv are not strings
        """
        PARAMS = ('client', 'path_to_store_csv')

        if not isinstance(kwargs.get('paths', list()), list):
            raise SDKException('Instance', '101')

        if any(param in kwargs for param in PARAMS) and \
                not all(isinstance(kwargs.get(param, None), str) for param in PARAMS):
            raise SDKException('Instance', '101')

        if not 'paths' in kwargs:
            kwargs['paths'] = ['/Files/', '/Objects/']

        request_json = self._restore_json(
            client=kwargs.get('client', self.proxy_client),
            restore_to_file_system=True,
            **kwargs
        )
        
        return self._process_restore_response(request_json)

    def restore_to_database(
            self,
            db_type,
            db_host_name,
            db_name,
            db_user_name,
            db_password,
            **kwargs
        ):
        """
        Runs object level restore to database and returns object of Job or Schedule class

        Args:
            db_type (str): Type of database out of 'POSTGRESQL' or 'SQLSERVER'

            db_host_name (str): Hostname of database server

            db_name (str): Database name where objects will be restored

            db_user_name (str): Username of database user

            db_password (str): Password of database user

            **kwargs (dict): Other restore options including
                {
                    paths (list[str]): List of files and objects to restore like
                                ['/Files/filename', '/Objects/object_name']
                                (Default is ['/Objects/'] which selects all objects for restore),

                    db_instance (str): Database instance for SQL Server,

                    db_port (int): Port of database server (Default is 5432 for POSTGRESQL and 1433 for SQLSERVER),

                    from_time (str): time to restore contents after like YYYY-MM-DD HH:MM:SS (Default is None),

                    to_time (str): time to restore contents before like YYYY-MM-DD HH:MM:SS (Default is None),

                    no_of_streams (int): Number of streams to use for restore (Default is 2),

                    path_to_store_csv (str): path to use as staging folder (Default is download cache path),

                    dependent_restore_level (int): restore children option (Default is 0)
                                                    0  -- No Children
                                                    1  -- Immediate Children
                                                    -1 -- All Children,

                    restore_parent_type (str): restore parents option (Default is 'NONE')
                                                    'NONE' -- No Parents
                                                    'ALL'  -- All Parents
                }

        Returns:
            object: Object of Job or Schedule class

        Raises:
            SDKException:
                if required parameters are not of the correct type

                if db_type is 'SQLSERVER' but db_instance is not given/ is not a string

                if paths is given but is not a list
        """
        PARAMS = (db_type, db_host_name,  db_name, db_user_name, db_password)

        if not isinstance(kwargs.get('paths', list()), list):
            raise SDKException('Instance', '101')

        if not all(isinstance(val, str) for val in PARAMS) and \
                (isinstance(kwargs.get('db_instance', None), str) != (db_type == 'SQLSERVER')):
            raise SDKException('Instance', '101')

        if not 'paths' in kwargs:
            kwargs['paths'] = ['/Objects/']

        request_json = self._restore_json(
            db_enabled=True,
            db_type=db_type,
            db_host=db_host_name,
            db_name=db_name,
            db_user_name=db_user_name,
            db_password=db_password,
            **kwargs
        )

        return self._process_restore_response(request_json)

    def restore_to_salesforce_from_database(self, **kwargs):
        """
        Runs restore to Salesforce from database and returns object of Job or Schedule class. For out of place restore,
        pass the client, instance and backupset parameters. If database parameters are not passed, sync db will be used.

        Args:
            **kwargs (dict): Other restore options including
                {
                    paths (list[str]): List of files and objects to restore like
                                ['/Files/filename', '/Objects/object_name']
                                (Default is ['/Files/', '/Objects/'] which selects all files and objects for restore),

                    client (str): Name of destination client (Default is source client),

                    instance (str): Name of destination instance (Default is source instance),

                    backupset (str): Name of destination backupset (Default is source backupset),

                    db_type (str): Type of database out of 'POSTGRESQL' or 'SQLSERVER',

                    db_host (str): Hostname of database server,

                    db_name (str): Database name where objects will be restored,

                    db_user_name (str): Username of database user,

                    db_password (str): Password of database user,

                    db_instance (str): Database instance for SQL Server,

                    db_port (int): Port of database server (Default is 5432 for POSTGRESQL and 1433 for SQLSERVER),

                    from_time (str): time to restore contents after like YYYY-MM-DD HH:MM:SS (Default is None),

                    to_time (str): time to restore contents before like YYYY-MM-DD HH:MM:SS (Default is None),

                    no_of_streams (int): Number of streams to use for restore (Default is 2),

                    path_to_store_csv (str): path to use as staging folder (Default is download cache path),

                    dependent_restore_level (int): restore children option (Default is 0)
                                                    0  -- No Children
                                                    1  -- Immediate Children
                                                    -1 -- All Children,

                    restore_parent_type (str): restore parents option (Default is 'NONE')
                                                    'NONE' -- No Parents
                                                    'ALL'  -- All Parents
                }

        Returns:
            object: Object of Job or Schedule class

        Raises:
            SDKException:
                if paths is given but is not a list

                if any database parameters are given but not all are present

                if database parameters are not all strings

                if db_type is 'SQLSERVER' but db_instance is not given/ is not a string

                if either client, instance or backupset are given but not all three are present

                if client, instance and backupset are not strings
        """
        DB_PARAMS = ('db_type', 'db_host', 'db_name', 'db_user_name', 'db_password')
        DEST_PARAMS = ('client', 'instance', 'backupset')

        if not isinstance(kwargs.get('paths', list()), list):
            raise SDKException('Instance', '101')

        if any(param in kwargs for param in DEST_PARAMS) and \
                not all(isinstance(kwargs.get(param, None), str) for param in DEST_PARAMS):
            raise SDKException('Instance', '101')

        if any(param in kwargs for param in DB_PARAMS):
            if not all(isinstance(kwargs.get(param, None), str) for param in DB_PARAMS) and \
                    (isinstance(kwargs.get('db_instance', None), str) != (kwargs['db_type'] == 'SQLSERVER')):
                raise SDKException('Instance', '101')
            kwargs['db_enabled'] = True

        if not 'paths' in kwargs:
            kwargs['paths'] = ['/Files/', '/Objects/']

        request_json = self._restore_json(
            restore_to_salesforce=True,
            restore_from_database=True,
            **kwargs
        )

        return self._process_restore_response(request_json)

    def restore_to_salesforce_from_media(self, **kwargs):
        """
        Runs restore to Salesforce from database and returns object of Job or Schedule class. For out of place restore,
        pass the client, instance and backupset parameters. If database parameters are not passed, sync db will be used
        as staging db.

        Args:
            **kwargs (dict): Other restore options including
                {
                    paths (list[str]): List of files and objects to restore like
                                ['/Files/filename', '/Objects/object_name']
                                (Default is ['/Files/', '/Objects/'] which selects all files and objects for restore),

                    client (str): Name of destination client (Default is source client),

                    instance (str): Name of destination instance (Default is source instance),

                    backupset (str): Name of destination backupset (Default is source backupset),

                    db_type (str): Type of database out of 'POSTGRESQL' or 'SQLSERVER',

                    db_host (str): Hostname of database server,

                    db_name (str): Database name where objects will be restored,

                    db_user_name (str): Username of database user,

                    db_password (str): Password of database user,

                    db_instance (str): Database instance for SQL Server,

                    db_port (int): Port of database server (Default is 5432 for POSTGRESQL and 1433 for SQLSERVER),

                    from_time (str): time to restore contents after like YYYY-MM-DD HH:MM:SS (Default is None),

                    to_time (str): time to restore contents before like YYYY-MM-DD HH:MM:SS (Default is None),

                    no_of_streams (int): Number of streams to use for restore (Default is 2),

                    path_to_store_csv (str): path to use as staging folder (Default is download cache path),

                    dependent_restore_level (int): restore children option (Default is 0)
                                                    0  -- No Children
                                                    1  -- Immediate Children
                                                    -1 -- All Children,

                    restore_parent_type (str): restore parents option (Default is 'NONE')
                                                    'NONE' -- No Parents
                                                    'ALL'  -- All Parents
                }

        Returns:
            object: Object of Job or Schedule class

        Raises:
            SDKException:
                if paths is given but is not a list

                if any database parameters are given but not all are present

                if database parameters are not all strings

                if db_type is 'SQLSERVER' but db_instance is not given/ is not a string

                if either client, instance or backupset are given but not all three are present

                if client, instance and backupset are not strings
        """
        DB_PARAMS = ('db_type', 'db_host', 'db_name', 'db_user_name', 'db_password')
        DEST_PARAMS = ('client', 'instance', 'backupset')

        if not isinstance(kwargs.get('paths', list()), list):
            raise SDKException('Instance', '101')

        if any(param in kwargs for param in DEST_PARAMS) and \
                not all(isinstance(kwargs.get(param, None), str) for param in DEST_PARAMS):
            raise SDKException('Instance', '101')

        if any(keyword in kwargs for keyword in DB_PARAMS):
            if not all(isinstance(kwargs.get(param, None), str) for param in DB_PARAMS) and \
                    (isinstance(kwargs.get('db_instance', None), str) != (kwargs['db_type'] == 'SQLSERVER')):
                raise SDKException('Instance', '101')
            kwargs['db_enabled'] = True

        if not 'paths' in kwargs:
            kwargs['paths'] = ['/Files/', '/Objects/']

        request_json = self._restore_json(
            restore_to_salesforce=True,
            **kwargs
        )

        return self._process_restore_response(request_json)

    def metadata_restore_to_file_system(self, **kwargs):
        """
        Runs metadata restore to file system and returns object of Job or Schedule class. For out of place restore,
        pass both client and path_to_store_csv parameters. By default, will restore to access node and download cache
        path.

        Args:
            **kwargs (dict): Other restore options including
                {                    
                    paths (list[str]): List of metadata components to restore like
                                ['/Metadata/unpackaged/objects/Account.object',
                                 '/Metadata/unpackaged/profiles/Admin.profile']
                                (Default is ['/Metadata/unpackaged/'] which selects all metdata components for restore),
                                
                    client (str): Name of destination client (Default is access node),
                    
                    path_to_store_csv (str): path on destination client to restore to (Default is download cache path),
                    
                    from_time (str): time to restore contents after like YYYY-MM-DD HH:MM:SS (Default is None),
                    
                    to_time (str): time to restore contents before like YYYY-MM-DD HH:MM:SS (Default is None),
                    
                    no_of_streams (int): Number of streams to use for restore (Default is 2),
                    
                    dependent_restore_level (int): restore children option (Default is 0)
                                                    0  -- No Children
                                                    1  -- Immediate Children
                                                    -1 -- All Children,
                                                    
                    restore_parent_type (str): restore parents option (Default is 'NONE')
                                                    'NONE' -- No Parents
                                                    'ALL'  -- All Parents
                }

        Returns:
            object: Object of Job or Schedule class

        Raises:
            SDKException:
                if paths is given but is not a list
    
                if client parameter is not given and the access node configured with this instance is a client group
                
                if either client or path_to_store_csv is given but not both are present
                
                if client or path_to_store_csv are not strings
        """
        PARAMS = ('client', 'path_to_store_csv')

        if not isinstance(kwargs.get('paths', list()), list):
            raise SDKException('Instance', '101')

        if any(param in kwargs for param in PARAMS) and \
                not all(isinstance(kwargs.get(param, None), str) for param in PARAMS):
            raise SDKException('Instance', '101')

        if not 'paths' in kwargs:
            kwargs['paths'] = ['/Metadata/unpackaged/']
        
        request_json = self._restore_json(
            restore_to_file_system=True,
            is_metadata_restore=True,
            **kwargs
        )
        return self._process_restore_response(request_json)

    def metadata_restore_to_salesforce(self, **kwargs):
        """
        Runs metadata restore to Salesforce and returns object of Job or Schedule class. For out of place restore,
        pass client, instance and backupset parameters.

        Args:
            **kwargs (dict): Other restore options including
                {
                    paths (list[str]): List of metadata components to restore like
                                ['/Metadata/unpackaged/objects/Account.object',
                                 '/Metadata/unpackaged/profiles/Admin.profile']
                                (Default is ['/Metadata/unpackaged/'] which selects all metdata components for restore),
                                
                    client (str): Name of destination client (Default is source client),
                    
                    instance (str): Name of destination instance (Default is source instance),
                    
                    backupset (str): Name of destination backupset (Default is source backupset),
                    
                    from_time (str): time to restore contents after like YYYY-MM-DD HH:MM:SS (Default is None),
                    
                    to_time (str): time to restore contents before like YYYY-MM-DD HH:MM:SS (Default is None),
                    
                    no_of_streams (int): Number of streams to use for restore (Default is 2),
                    
                    path_to_store_csv (str): path to use as staging folder (Default is download cache path),
                    
                    dependent_restore_level (int): restore children option (Default is 0)
                                                    0  -- No Children
                                                    1  -- Immediate Children
                                                    -1 -- All Children,
                                                    
                    restore_parent_type (str): restore parents option (Default is 'NONE')
                                                    'NONE' -- No Parents
                                                    'ALL'  -- All Parents
                }

        Returns:
            object: Object of Job or Schedule class

        Raises:
            SDKException:
                if paths is given but is not a list
                
                if either client, instance or backupset are given but not all three are present

                if client, instance and backupset are not strings
        """
        DEST_PARAMS = ('client', 'instance', 'backupset')

        if not isinstance(kwargs.get('paths', list()), list):
            raise SDKException('Instance', '101')

        if any(param in kwargs for param in DEST_PARAMS) and \
                not all(isinstance(kwargs.get(param, None), str) for param in DEST_PARAMS):
            raise SDKException('Instance', '101')

        if not 'paths' in kwargs:
            kwargs['paths'] = ['/Metadata/unpackaged/']
        
        request_json = self._restore_json(
            restore_to_salesforce=True,
            is_metadata_restore=True,
            **kwargs
        )
        
        return self._process_restore_response(request_json)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server OracleVM.

OracleVMVirtualServerSubclient is the only class defined in this file.

OracleVMVirtualServerSubclient: Derived class from VirtualServerSubClient  Base class,
                            representing a OracleVM Subclient, and
                            to perform operations on that Subclient

OracleVMVirtualServerSubclient:

    __init__(,backupset_object, subclient_name, subclient_id)--  initialize object of FusionCompute
                                                                             subclient object
                                                                                 associated with
                                                                        the VirtualServer subclient

    full_vm_restore_out_of_place()                  --  restores the VM  specified in
                                                     to the specified client, at the
                                                        specified destination location

    full_vm_restore_in_place()              --  restores the VM specified by the
                                                    user to the same location
"""

from ..vssubclient import VirtualServerSubclient
from ...exception import SDKException


class OracleVMVirtualServerSubclient(VirtualServerSubclient):
    """Derived class from VirtualServerSubclient Base class.
           This represents a OracleVM virtual server subclient,
           and can perform restore operations on only that subclient.

        """
    def __init__(self, backupset_object, subclient_name, subclient_id=None):
        """Initialize the Instance object for the given Virtual Server instance.
        Args
        class_object (backupset_object, subclient_name, subclient_id)  --  instance of the
                                         backupset class, subclient name, subclient id

        """

        super(OracleVMVirtualServerSubclient, self).__init__(
            backupset_object, subclient_name, subclient_id)
        self.diskExtension = [".img"]

    def full_vm_restore_in_place(
            self,
            vm_to_restore=None,
            destination_client=None,
            overwrite=True,
            power_on=True,
            disk_option='Original',
            transport_mode='Auto',
            copy_precedence=0,
            **kwargs):
        """Restores the FULL Virtual machine specified in the input list
            to the location same as the actual location of the VM in VCenter.

            Args:
                vm_to_restore       (list)     --  provide the VM name to
                                                   restore
                                                   default: None

                destination_client (str)  -- proxy client to be used for
                                                    restore
                                                    default: proxy added in
                                                    subclient

                overwrite           (bool)     --  overwrite the existing VM
                                                   default: True

                power_on            (bool)     --  power on the  restored VM
                                                   default: True

                disk_option       (str) -- disk provisioning for the
                                                  restored vm
                                                  default: 0 which is equivalent
                                                  to Original

                transport_mode    (str) -- transport mode that need to be
                                                  used

                copy_precedence     (int)      --  copy precedence value
                                                   default: 0

                **kwargs                         : Arbitrary keyword arguments Properties as of
                                                     full_vm_restore_in_place
                    eg:
                    v2_details          (dict)       -- details for v2 subclient
                                                    eg: check clients.vmclient.VMClient._child_job_subclient_details


            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        restore_option = {"v2_details": kwargs.get("v2_details", None)}

        if copy_precedence:
            restore_option["copy_precedence_applicable"] = True

        # set attr for all the option in restore xml from user inputs
        self._set_restore_inputs(
            restore_option,
            unconditional_overwrite=overwrite,
            power_on=power_on,
            copy_precedence=copy_precedence,
            vm_to_restore=self._set_vm_to_restore(vm_to_restore),
            volume_level_restore=1,
            client_name=destination_client,
            in_place=True,
            transport_mode=self._transport_mode.get(transport_mode.replace(" ", "").lower(),
                                                    self._transport_mode["auto"]),
            disk_option=self._get_disk_provisioning_value(disk_option),
        )

        request_json = self._prepare_fullvm_restore_json(restore_option)
        return self._process_restore_response(request_json)

    def full_vm_restore_out_of_place(
            self,
            vm_to_restore=None,
            restored_vm_name=None,
            disk_name_prefix=None,
            virtualization_client=None,
            destination_client=None,
            repository=None,
            overwrite=True,
            power_on=True,
            server=None,
            copy_precedence=0,
            disk_provisioning='Original',
            **kwargs):
        """Restores the FULL Virtual machine specified in the input list
            to the provided vcenter client along with the ESX and the datastores.
            If the provided client name is none then it restores the Full Virtual
            Machine to the source client and corresponding ESX and datastore.

            Args:
                vm_to_restore     (string)  --  provide the VM name to restore
                                              {"name_of_vm_to_restore":
                                              "new_name_of_restored_vm"}
                                              default: {}
                restored_vm_name   (string)  -- name of the new VM that should
                                                restored.

                disk_name_prefix    (string) -- new name for the disks while restoring
                                                the VM

                virtualization_client   (str) -- name of the Oracle
                                                    virtualization
                                                    client that needs to be
                                                    restored

                destination_client    (str) -- name of the proxy that
                                                    should be used during
                                                    restore

                repository         (str) -- datastore where the
                                                  restored VM should be located
                                                  restores to the source VM
                                                  datastore if this value is
                                                  not specified

                overwrite         (bool)       -- overwrite the existing VM
                                                  default: True

                power_on          (bool)       -- power on the  restored VM
                                                  default: True

                server          (str) -- destination cluster or  host
                                                    restores to the source VM
                                                    esx if this value is not
                                                    specified

                copy_precedence   (int)        -- copy precedence value
                                                  default: 0

                disk_provisioning       (str) -- disk provisioning for the
                                                  restored vm default: 0
                                                  which is equivalent
                                                  to Original

                **kwargs                         : Arbitrary keyword arguments Properties as of
                                                     full_vm_restore_out_of_place
                    eg:
                    v2_details          (dict)       -- details for v2 subclient
                                                    eg: check clients.vmclient.VMClient._child_job_subclient_details



            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        restore_option = {"v2_details": kwargs.get("v2_details", None)}

        if vm_to_restore and not isinstance(vm_to_restore, str):
            raise SDKException('Subclient', '101')

        if copy_precedence:
            restore_option["copy_precedence_applicable"] = True

        if restored_vm_name:
            if not (isinstance(vm_to_restore, str) or
                    isinstance(restored_vm_name, str)):
                raise SDKException('Subclient', '101')
            restore_option['restore_new_name'] = restored_vm_name

        if vm_to_restore:
            vm_to_restore = [vm_to_restore]

        self._set_restore_inputs(
            restore_option,
            vm_to_restore=self._set_vm_to_restore(vm_to_restore),
            unconditional_overwrite=overwrite,
            power_on=power_on,
            disk_option=self._get_disk_provisioning_value(disk_provisioning),
            restore_new_name=restored_vm_name,
            copy_precedence=copy_precedence,
            volume_level_restore=1,
            client_name=destination_client,
            vcenter_client=virtualization_client,
            esx_host=server,
            datastore=repository,
            in_place=False,
            disk_name_prefix=disk_name_prefix
        )

        request_json = self._prepare_fullvm_restore_json(restore_option)
        return self._process_restore_response(request_json)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for managing global filters for this commcell

GlobalFilters and GlobalFilter are the only classes defined in this file

GlobalFilters: Class for managing global filters for this commcell

GlobalFilter: Class to represent one agent specific global filter

GlobalFilters:
    __init__()                      --  initializes global filter class object

    __repr__()                      --  returns the string for the instance of the GlobalFilter
                                            class

    get()                           --  returns the GlobalFilter object for specified filter name


GlobalFilter:
    __init__()                      --  initializes global filter object

    __repr__()                      --  returns string representing this class

    _get_global_filters()           --  gets the global filters associated with commcell
                                            for specified filter

    _initialize_global_filters()    --  initializes GlobalFilter class objects

    _update()                       --  updates the global filters list on commcell

    content()                       --  returns the list of filters associated with this agent

    add()                           --  adds the specified filter to global list

    overwrite()                     --  overwrites existing global list with specified

    delete_all()                    --  removes all the filters from global filters list

    refresh()                       --  refresh the properties of the global filter

"""

from .exception import SDKException


class GlobalFilters(object):
    """Class for managing global filters for this commcell"""

    def __init__(self, commcell_object):
        """Initializes global filter object

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the GlobalFilter class
        """
        self._commcell_object = commcell_object

        self._global_filter_dict = {
            "WINDOWS": 'windowsGlobalFilters',
            "UNIX": 'unixGlobalFilters',
            "NAS":  'nasGlobalFilters'
        }

    def __repr__(self):
        """Representation string for the instance of the GlobalFilter class."""
        o_str = "GlobalFilter class instance for CommServ '{0}'".format(
            self._commcell_object.commserv_name
        )
        return o_str

    def get(self, filter_name):
        """Returns the global filter agent object for specified filter name

            Args:
                filter_name     (str)   -- Global filter name for which the object is to be created
                    Accepted values: WINDOWS/ UNIX/ NAS

            Returns:
                object - GlobalFilter object for specified global filter

            Raises:
                SDKException:
                    if data type of input is invalid

                    if specified global filter doesn't exist
        """
        if not isinstance(filter_name, str):
            raise SDKException('GlobalFilter', '101')

        if filter_name.upper() not in self._global_filter_dict:
            raise SDKException(
                'GlobalFilter', '102', 'Invalid Global Filter name {0}'.format(filter_name)
            )

        return GlobalFilter(
            self._commcell_object,
            filter_name.upper(),
            self._global_filter_dict[filter_name.upper()]
        )


class GlobalFilter(object):
    """Class to represent any one particular agent global filter"""

    def __init__(self, commcell_object, filter_name, filter_key):
        """Initializes global filter object

            Args:
                commcell_object     (object)    -- commcell object

                agent_key           (str)       --  agent key that shall be used in requests
        """
        self._filter_name = filter_name
        self._filter_key = filter_key
        self._commcell_object = commcell_object
        self._GLOBAL_FILTER = self._commcell_object._services['GLOBAL_FILTER']
        self._content = []

        self.refresh()

    def __repr__(self):
        """String representation of the instance of this class."""
        return "Global Filter object for: {0}".format(self._filter_name)

    def _get_global_filters(self):
        """Returns the global filters associated with this commcell"""
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._GLOBAL_FILTER
        )

        if flag:
            if response.json():
                return response.json()
            else:
                return {}
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def _initialize_global_filters(self):
        """Initializes global filters"""
        global_filters = self._get_global_filters()

        self._content = []

        if self._filter_key in global_filters:
            self._content = global_filters[self._filter_key]

    def _update(self, op_type, filters_list):
        """Updates the global filters list on tise commcell

            Args:
                op_type         (dict)  --  operation type to be performed
                        Accepted values: ADD/ OVERWRITE/ DELETE

                filters_list    (list)  --  list of filters to be associated

            Raises:
                SDKException:
                    if failed to update global filter content

                    if response received is empty

                    if response is not success
        """
        op_dict = {
            "ADD": 1,
            "OVERWRITE": 1,
            "DELETE": 3
        }

        request_json = {
            self._filter_key: {
                "opType": op_dict[op_type],
                "filters": filters_list
            }
        }

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._GLOBAL_FILTER, request_json
        )

        self.refresh()

        if flag:
            if response.json() and 'error' in response.json():
                if 'errorCode' in response.json()['error']:
                    error_code = int(response.json()['error']['errorCode'])

                    if error_code != 0:
                        raise SDKException(
                            'GlobalFilter', '102', 'Failed to update global filters'
                        )
                else:
                    raise SDKException('Response', '102')
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    @property
    def content(self):
        """Treats filter content as read-only property"""
        return self._content

    def add(self, filters_list):
        """Adds the filters list to the specified agent global filters list

            Args:
                filters_list    (list)  --  list of filters to be added to this agent

            Raises:
                SDKException:
                    if data type of input is invalid

                    if failed to update global filter content

                    if response received is empty

                    if response is not success
        """
        if not isinstance(filters_list, list):
            raise SDKException('GlobalFilter', '101')

        self._update("ADD", filters_list + self.content)

    def overwrite(self, filters_list):
        """Overwrites the existing filters list with given filter list

            Args:
                filters_list    (list)  --  list of filters to be replaced with existing

            Raises:
                SDKException:
                    if data type of input is invalid

                    if failed to update global filter content

                    if response received is empty

                    if response is not success
        """
        if not isinstance(filters_list, list):
            raise SDKException('GlobalFilter', '101')

        self._update("OVERWRITE", filters_list)

    def delete_all(self):
        """Deletes all the filters from given agent filters list

            Raises:
                SDKException:
                    if failed to update global filter content

                    if response received is empty

                    if response is not success
        """
        self._update("OVERWRITE", [""])

    def refresh(self):
        """Refresh the properties of the GlobalFilter."""
        self._initialize_global_filters()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on all types of Policies associated with the Commcell.

Policies:   Class for representing all types of Policies associated with the Commcell

"""

from __future__ import unicode_literals

from .exception import SDKException

from .policies.configuration_policies import ConfigurationPolicies
from .policies.storage_policies import StoragePolicies
from .policies.schedule_policies import SchedulePolicies


class Policies:
    """Class for getting all the all the policies associated with the commcell."""

    def __init__(self, commcell_object):
        """Initialize object of the Policies class.

            Args:
                commcell_object     (object)    --  instance of the Commcell class

            Returns:
                object  -   instance of the Policies class

        """
        self._commcell_object = commcell_object
        self.refresh()

    def __repr__(self):
        """Representation string for the instance of the Policies class."""
        return "Policies class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def refresh(self):
        """Refresh all the Policies associated with the Commcell."""
        self._configuration_policies = None
        self._storage_policies = None
        self._schedule_policies = None

    @property
    def configuration_policies(self):
        """Returns the instance of the ConfigurationPolicies class."""
        if self._configuration_policies is None:
            self._configuration_policies = ConfigurationPolicies(self._commcell_object)

        return self._configuration_policies


    @property
    def storage_policies(self):
        """Returns the instance of the StoragePolicies class."""
        if self._storage_policies is None:
            self._storage_policies = StoragePolicies(self._commcell_object)

        return self._storage_policies


    @property
    def schedule_policies(self):
        """Returns the instance of the SchedulePolicies class."""
        if self._schedule_policies is None:
            self._schedule_policies = SchedulePolicies(self._commcell_object)

        return self._schedule_policies



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server Fusion Compute Instance.

FusionComputeInstance is the only class defined in this file.

FusionComputeInstance: Derived class from VirtualServer  Base class, representing a
                           Fusion Compute instance, and to perform operations on that instance

HyperVInstance:

    __init__(agent_object,instance_name,instance_id)    -- initialize object of FusionCompute
                                                                Instance object associated with the
                                                                        VirtualServer Instance


    _get_instance_properties()                          --  VirtualServer Instance class method
                                                            overwritten to get Fusion Compute
                                                            Specific instance properties as well

    _set_instance_properties()                          --  Fusion Compute Instance class method
                                                                to set Fusion Compute
                                                                Specific instance properties


"""

from ..vsinstance import VirtualServerInstance


class FusionComputeInstance(VirtualServerInstance):
    """Class for representing a Fusion Compute of the Virtual Server agent."""

    def __init__(self, agent, instance_name, instance_id=None):
        """Initialize the Instance object for the given Virtual Server instance.

            Args:
                class_object (agent_object,instance_name,instance_id)  --  instance of the
                                                                                Agent class,
                                                                                instance name,
                                                                                instance id

        """

        self._vendor_id = 14
        self._server_name = []
        self._vmwarvendor = None
        self._server_host_name = []
        super(FusionComputeInstance, self).__init__(agent, instance_name, instance_id)

    def _get_instance_properties(self):
        """
        Get the properties of this instance

        Raise:
            SDK Exception:
                if response is not empty
                if response is not success
        """

        super(FusionComputeInstance, self)._get_instance_properties()
        if "vmwareVendor" in self._virtualserverinstance:
            self._vmwarvendor = self._virtualserverinstance['vmwareVendor']['virtualCenter']

            self._server_name.append(self._instance['clientName'])

            self._server_host_name.append(self._vmwarvendor["domainName"])

    def _get_instance_properties_json(self):
        """get the all instance related properties of this subclient.

          Returns:
               dict - all instance properties put inside a dict

        """
        instance_json = {
            "instanceProperties": {
                "isDeleted": False,
                "instance": self._instance,
                "instanceActivityControl": self._instanceActivityControl,
                "virtualServerInstance": {
                    "vsInstanceType": self._virtualserverinstance['vsInstanceType'],
                    "associatedClients": self._virtualserverinstance['associatedClients'],
                    "vmwareVendor": self._virtualserverinstance['vmwareVendor']
                }
            }
        }
        return instance_json

    @property
    def server_host_name(self):
        """return the Fusion compute VRM  associated with the PseudoClient"""
        return self._server_host_name

    @property
    def server_name(self):
        """getter for the domain name in the vmware vendor json"""
        return self._server_name

    @property
    def _user_name(self):
        """getter for the username from the vmware vendor json"""
        return self._vmwarvendor.get("userName", "")



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""
Main file for getting failover group related information

FailoverGroups and FailoverGroup are 2 classes defined in this file.

FailoverGroups:     Class for getting information of all failover groups in the commcell

FailoverGroup:      Class for a failover group that gives us all the live sync pairs associated to in,
                    with addition to the clients/hypervisors associated


FailoverGroups:
    FailoverGroupSourceTypes                    --  Enum to represent all type of sources for failover groups
    FailoverGroupTypes                          --  Enum to represent all types of failover groups
    DRReplicationTypes                          --  Enum to represent all live sync types
    __init__(commcell_object)                   --  Initialize the object of failovergroups class for commcell

    __str__()                                   --  Returns the list of all failover groups

    __repr__()                                  --  Returns the string for the instance of the FailoverGroups class

    has_failover_group(failover_group_name)     --  Checks if failover group exists with the given name

    get(failover_group_name)                    --  Returns the FailoverGroup class object of the given name

    refresh()                                   --  Refresh all failover groups created on the commcell

    #### internal methods ###
    _get_failover_groups()                      --  Internal call to get information of all failover groups in commcell

    ##### properties ######
    failover_groups                             --  Returns the dictionary of all failover groups and their info


FailoverGroup:
    __init__(commcell_object,
            failover_group_name)                --  Initialize object of FailoverGroup with the given name

    __repr__()                                  --  Returns the name of the failover group for the object

    __str__()                                   --  Returns the name of the all VM pairs for the failover group

    refresh()                                   --  Refresh the failover group properties

    ##### internal methods #####
    _get_failover_group_dict()                  --  Gets the failover group information from FailoverGroups class

    _get_failover_group_properties()            --  Get the failover group properties

    ##### properties #####
    failover_group_id                           --  The ID of the failover group

    failover_group_name                         --  The name of the failover group

    replication_type                            --  The DRReplicationTypes key for replication of failover group

    group_type                                  --  The FailoverGroupTypes for operation of failover group

    source_type                                 --  The FailoverGroupSourceTypes of source of failover group

    is_client_group                             --  Whether the VM pairs are part of a client group or not

    replication_pairs                           --  The ReplicationPairs class for failover group

    vm_pair_ids                                 --  The ID of the replication pairs

    vm_pairs                                    --  Returns the live sync pair objects for each VM pair of the group
                                                        as a mapping of source VM name and VM pair object

    replication_groups                          --  The names of all replication groups associated
                                                    to the failover group

    source_client                               --  The client object of the source client

    source_agent                                --  The agent object of the source client

    source_instance                             --  The instance object of the source client

    destination_client                          --  The client object of the destination client

    destination_agent                           --  The agent object of the destination client

    destination_instance                        --  The instance object of the destination instance

    is_approval_required                        --  Whether the approval is set in failover group or not

    user_for_approval                           --  Returns the name of the user set in failover group
                                                        for approval

"""
from enum import Enum
from ..exception import SDKException
from .replication_pairs import ReplicationPairs


class FailoverGroups:
    """Class for getting all the failover groups in commcell."""

    class FailoverGroupSourceTypes(Enum):
        BACKUP = 0
        REPLICATION = 1
        TEMPLATES = 2

    class FailoverGroupTypes(Enum):
        """ Enum to map Failover Group Types to integers"""
        LIVE_MOUNT = 1
        LIVE_SYNC = 2
        RESTORE = 4
        LIVE_RECOVERY = 8
        FAILOVER = 16
        VIRTUAL_LAB = 32
        ORACLE_EBS_APP = 64
        GENERIC_ENTERPRISE_APP = 128
        TEST_FAILOVER = 256

    class DRReplicationTypes(Enum):
        """ Enum to map replication types to replication groups/failover groups"""
        LIVE_SYNC = 0
        LIVE_SYNC_DIRECT = 1
        LIVE_SYNC_IO = 2
        SNAP_ARRAY = 3

    def __init__(self, commcell_object):
        """Initialize object of the Failover groups
            Args:
                commcell_object (Commcell)  --  instance of the Commcell class
        """
        self._commcell_object = commcell_object
        self._services = commcell_object._services

        self._failover_groups = None

        self.refresh()

    def __str__(self):
        """Representation string consisting of all failover groups in a formatted output
            Returns:
                str - string of all the failover groups
        """
        representation_string = (f'{"S. No.":^5}\t'
                                 f'{"Failover Group Id":^20}\t'
                                 f'{"Failover Group":^20}\n\n')

        for index, failover_group in enumerate(self._failover_groups):
            sub_str = (f'{index + 1:^5}\t'
                       f'{self._failover_groups[failover_group]:20}\t'
                       f'{failover_group:20}\n')
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the FailoverGroups class."""
        return f"Failover Groups for Commserv: '{self._commcell_object.commserv_name}'"

    def has_failover_group(self, failover_group_name):
        """Checks if failover group exists or not

            Args:
                failover_group_name (str)  --  name of the failover group

            Returns:
                bool - boolean output whether failover group exists or not

            Raises:
                SDKException:
                    if proper inputs are not provided
        """
        if not isinstance(failover_group_name, str):
            raise SDKException('FailoverGroup', '101')

        return self.failover_groups and failover_group_name.lower() in self.failover_groups

    def get(self, failover_group_name):
        """Returns a failover group object of the specified failover group name.

            Args:
                failover_group_name (str)  --  name of the failover group

            Returns:
                object - instance of the FailoverGroup class for the given failover group name

            Raises:
                SDKException:
                    if proper inputs are not provided
                    If Failover group doesnt exists with given name
        """
        if not isinstance(failover_group_name, str):
            raise SDKException('FailoverGroup', '101')
        failover_group_name = failover_group_name.lower()
        if not self.has_failover_group(failover_group_name):
            raise SDKException('FailoverGroup', '103')
        return FailoverGroup(self._commcell_object, failover_group_name)

    @property
    def failover_groups(self):
        """ return all failover groups
        Args:

        Returns: (dict) All the failover groups in the commcell
                eg:
                {
                     "failover_group_name1": {id: '1', 'type': VSA_PERIODIC, 'operation_type': FAILOVER},
                     "failover_group_name2": {id: '2', 'type': VSA_CONTINUOUS, 'operation_type': FAILOVER}
                }
        Raises:
        """
        return self._failover_groups

    def _get_failover_groups(self):
        """REST API call for all the failover groups in the commcell.
            Args:

            Returns:
                dict - consists of all failover groups
                    {
                        "failover_group_name1": {
                            'id': '1',
                            'type': LIVE_SYNC,
                            'operation_type': FAILOVER,
                            'source_type': REPLICATION
                        },
                        "failover_group_name2": {
                            'id': '2',
                            'type': SNAP_ARRAY,
                            'operation_type': TEST_FAILOVER
                            'source_type': REPLICATION
                        }
                    }

            Raises:
                SDKException:
                    if response is empty
                    if response is not success
        """
        failover_groups = {}

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', (self._services['FAILOVER_GROUPS']
                    if self._commcell_object.commserv_version > 30
                    else self._services['DR_GROUPS']))
        if flag:
            if 'vApp' in response.json():
                for failover_group in response.json().get('vApp', []):
                    failover_group_id = str(failover_group.get('vAppEntity', {}).get('vAppId'))
                    failover_group_name = failover_group.get('vAppEntity', {}).get('vAppName', '').lower()
                    operation_type = self.FailoverGroupTypes(int(failover_group.get('operationType', 16)))
                    replication_type = self.DRReplicationTypes(int(failover_group.get('replicationType', 0)))
                    source_type = self.FailoverGroupSourceTypes(int(failover_group.get('source', 1)))
                    failover_groups[failover_group_name] = {
                        'id': failover_group_id,
                        'operation_type': operation_type,
                        'type': replication_type,
                        'source_type': source_type
                    }
                return failover_groups
            raise SDKException('Response', '102')

        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """ Refresh the failover groups created in the commcell.
        Args:

        Returns:

        Raises:

        """
        self._failover_groups = self._get_failover_groups()


class FailoverGroup:
    """ Class for representing a failover group """

    def __init__(self, commcell_object, failover_group_name):
        """Initialise the FailoverGroup object for the given group name
            Args:
                commcell_object (Commcell)      --  instance of the Commcell class
                failover_group_name (str)       --  name of the failover group
        """
        self._commcell_object = commcell_object
        self._services = commcell_object._services
        self._failover_group_properties = None

        self._failover_group_name = failover_group_name.lower()
        self._failover_group_dict = self._get_failover_group_dict()

        self._source_client = None
        self._destination_client = None
        self._destination_agent = None
        self._source_instance = None
        self._destination_instance = None

        self._replication_pairs = None

        self.refresh()

    def __repr__(self):
        """String representation of the instance of the failover group"""
        return f'FailoverGroup class instance for {self._failover_group_name}'

    def __str__(self):
        """Strings showing all VM pairs of the failover group in a formatted output
            Returns:
                str - string of all VM pairs
        """
        representation_string = f'{"Pair Id":^5}\t{"Source VM":^20}\t{"Destination VM":^20}\n\n'

        for source_vm in self.vm_pairs:
            sub_str = (f'{self.vm_pairs[source_vm].vm_pair_id:^5}\t'
                       f'{source_vm:20}\t'
                       f'{self.vm_pairs[source_vm].destination_vm:20}'
                       f'\n')
            representation_string += sub_str

        return representation_string.strip()

    def _get_failover_group_dict(self):
        """Get the failover group's basic information from FailoverGroups object for Commcell"""
        fgs_obj = FailoverGroups(self._commcell_object)
        return fgs_obj.failover_groups.get(self._failover_group_name)

    def _get_failover_group_properties(self):
        """ Gets failover group properties
            Args:

            Returns: Gets the failover group properties dict

            Raises:
                SDKException:
                    if response is empty

                    if response is not success
        """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET',
            (self._services['GET_FAILOVER_GROUP']
             if self._commcell_object.commserv_version > 30
             else self._services['GET_DR_GROUP']) % str(self.failover_group_id)
        )
        if flag:
            if 'vApp' in response.json():
                return response.json().get('vApp', [{}])[0]
            raise SDKException('Response', '102')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """ Refresh the failover group properties """
        self._failover_group_properties = self._get_failover_group_properties()
        self.replication_pairs.refresh()

    @property
    def failover_group_id(self):
        """Returns: (str) The ID of the failover group"""
        return self._failover_group_dict.get('id')

    @property
    def failover_group_name(self):
        """Returns: (str) The name of the failover group"""
        return self._failover_group_name

    @property
    def replication_type(self):
        """Returns: (DRReplicationTypes) The type of replication"""
        return self._failover_group_dict.get('type')

    @property
    def group_type(self):
        """Returns: (FailoverGroupTypes) The type of failover group"""
        return self._failover_group_dict.get('operation_type')

    @property
    def source_type(self):
        """Returns: (FailoverGroupSourceTypes) The type of failover group's source"""
        return self._failover_group_dict.get('source_type')

    @property
    def is_client_group(self):
        """Returns: (bool) Whether this failover group has a client group or not"""
        return self._failover_group_properties.get('isClientGroup')

    @property
    def replication_pairs(self):
        """
        Returns: (ReplicationPairs) Returns the ReplicationPairs object that belongs to this failover group
        Note: Implemented only for live sync failover groups
        """
        if not self._replication_pairs:
            if self.replication_type == FailoverGroups.DRReplicationTypes.LIVE_SYNC:
                self._replication_pairs = ReplicationPairs(self._commcell_object,
                                                           failover_group_id=self.failover_group_id)
        return self._replication_pairs

    @property
    def vm_pair_ids(self):
        """Returns: (List[str]) Returns the VM pair IDs that belong to this failover group"""
        return list(self.replication_pairs.replication_pairs)

    @property
    def vm_pairs(self):
        """
        Returns: (dict) The list of all live sync VM pairs
            eg:
                {
                    <source_vm1>: <Replication_pair_obj1>,
                    <source_vm2>: <Replication_pair_obj2>
                }
        """
        replication_pair_objects = {}
        for replication_pair_id, replication_pair_dict in self.replication_pairs.replication_pairs.items():
            source_vm_name = replication_pair_dict.get('source_vm')
            replication_pair_object = self.replication_pairs.get(replication_id=replication_pair_id)
            replication_pair_objects[source_vm_name] = replication_pair_object
        return replication_pair_objects

    @property
    def replication_groups(self):
        """Returns: (list) The list of all replication group names"""
        group_names = {vm_pair.replication_group_name for vm_pair in self.vm_pairs.values()}
        return list(group_names)

    @property
    def source_client(self):
        """Returns: (Client) The client object for the failover group's source hypervisor"""
        if not self._source_client:
            client_name = self._failover_group_properties.get('selectedEntities', [{}])[0].get('entityName', '')
            self._source_client = self._commcell_object.clients.get(client_name)
        return self._source_client

    @property
    def source_agent(self):
        """Returns: (Agent) The agent object for the source hypervisor"""
        return self.source_instance._agent_object

    @property
    def source_instance(self):
        """Returns: (Instance) The instance object for the source hypervisor"""
        if not self._source_instance:
            instance_id = self._failover_group_properties.get('selectedEntities', [{}])[0].get('instanceId', '')
            for agent_name in self.source_client.agents.all_agents:
                agent_object = self.source_client.agents.get(agent_name)
                for instance_name, inst_id in agent_object.instances.all_instances.items():
                    if inst_id == str(instance_id):
                        self._source_instance = agent_object.instances.get(instance_name)
                        return self._source_instance

        return self._source_instance

    @property
    def destination_client(self):
        """Returns: (Client) The client object for the failover group's destination hypervisor"""
        return self.destination_agent._client_object

    @property
    def destination_agent(self):
        """Returns: (Agent) The agent object for the destination hypervisor"""
        return self.destination_instance._agent_object

    @property
    def destination_instance(self):
        """Returns: (Instance) The instance object for the destination hypervisor"""
        if not self._destination_client:
            vm_pair = list(self.vm_pairs.values())[0]
            self._destination_instance = (vm_pair._subclient_object
                                          ._backupset_object._instance_object)
        return self._destination_instance

    @property
    def is_approval_required(self):
        """Returns bool:
            true : if approval set in failover group
            false: if approval not set in failover group
        """
        return self._failover_group_properties.get('approvalRequired')

    @property
    def user_for_approval(self):
        """Returns: user name set in failover group"""
        user_name = (self._failover_group_properties.get('usersForApproval', [{}])[0]
                     .get('userEntity', {}).get('userName', ''))
        return user_name
